/**
 * Avahi profile scanner plugin
 *
 * By Jim Ramsay <i.am@jimramsay.com>
 *
 * Released under the GPL.
 *
 */
#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>
#include <libmpd/debug_printf.h>
#include <gmpc/plugin.h>
#include <gmpc/gmpc-profiles.h>

#include <avahi-client/client.h>
#include <avahi-client/lookup.h>

#include <avahi-glib/glib-watch.h>
#include <avahi-glib/glib-malloc.h>
#include <avahi-common/domain.h>
#include <avahi-common/error.h>

#define SERVICE_TYPE "_mpd._tcp"

static void avahi_domain_changed(void);
static const char* avahi_get_default_domain(void);
static void avahi_add_service( const char *name, const char *host, int port );
static void avahi_del_service( const char *name );
static const char* avahi_get_browse_domain(void);

/**** This section is for the gmpc plugin part ****/

void avahi_init(void);

void avahi_pref_construct(GtkWidget *container);
void avahi_pref_destroy(GtkWidget *container);

//int avahi_menu(GtkWidget *menu);
//Profile* avahi_next(Profile* current);

void avahi_set_enabled(int enabled);
int avahi_get_enabled(void);

void avahi_update_profiles_menu(void);

static GtkWidget *pref_vbox;
gmpcPrefPlugin avahi_prefs = {
	.construct = avahi_pref_construct,
	.destroy = avahi_pref_destroy
};

/*gmpcProfilesPlugin avahi_profiles = {
	.category = "Avahi",
	.next = avahi_next,
};
*/
int plugin_api_version = PLUGIN_API_VERSION;
gmpcPlugin plugin = {
	.name = "Avahi Zeroconf",
	.version = {0,15,0},
	.plugin_type = GMPC_PLUGIN_DUMMY,
	.init = avahi_init,
	.pref = &avahi_prefs,
//	.profiles = &avahi_profiles,
	.get_enabled = avahi_get_enabled,
	.set_enabled = avahi_set_enabled,
};

/**
 * Get/Set enabled
 */
int avahi_get_enabled()
{
	return cfg_get_single_value_as_int_with_default(config, "avahi-profiles", "enable", TRUE);
}
void avahi_set_enabled(int enabled)
{
	cfg_set_single_value_as_int(config, "avahi-profiles", "enable", enabled);
}

/**
 * Profiles
 */
typedef struct ProfileIter_s {
	struct ProfileIter_s *next, *prev;
	//Profile* profile;
} ProfileIter; 

static ProfileIter* profileList = NULL;

void avahi_add_service( const char *name, const char *host, int port )
{
	if(gmpc_profiles_has_profile(gmpc_profiles, name))
	{	
		/* update host/port */
		if(strcmp(gmpc_profiles_get_hostname(gmpc_profiles,name), host))
		{
			gchar *cname = g_strdup_printf("%s (%s)", name,host);
			gmpc_profiles_set_name(gmpc_profiles, name,cname);
			g_free(cname);

			gmpc_profiles_set_hostname(gmpc_profiles, name,host);
		}
		if(gmpc_profiles_get_port(gmpc_profiles,name) != port)
		{
			gmpc_profiles_set_port(gmpc_profiles, name,port);
		}
	}
	else
	{
		/* create new */
		gchar *id = gmpc_profiles_create_new_item(gmpc_profiles, name);
		gchar *cname = g_strdup_printf("%s (%s)", name,host);
		gmpc_profiles_set_name(gmpc_profiles,id, cname);
		g_free(cname);
		gmpc_profiles_set_hostname(gmpc_profiles, id,host);
		gmpc_profiles_set_port(gmpc_profiles, id,port);
	}
/*	Profile *newProfile = g_malloc( sizeof(Profile) );

	// Thanks to collision avoidance, the name is guaranteed to be okay and
	// therefor a good id
	newProfile->id = g_strdup( name );
	newProfile->category = avahi_profiles.category;
	newProfile->name = g_strdup( name );
	newProfile->hostname = g_strdup( host );
	newProfile->portnumber = port;
	newProfile->password = NULL;
	newProfile->useauth = FALSE;

	// Only the password can be edited
	newProfile->flags = GMPC_PROFILES_CAN_EDIT_PASSWORD | GMPC_PROFILES_CAN_EDIT_USEAUTH;

	ProfileIter *newIter = g_malloc( sizeof(ProfileIter) );
	newIter->next = profileList;
	newIter->profile = newProfile;
	newIter->prev = NULL;
	newProfile->data = newIter;

	profileList = newIter;
	if( newIter->next )
		newIter->next->prev = newIter;

	gmpc_profile_added( &plugin, newIter->profile );
*/
}

void avahi_free_service( /*ProfileIter* it */)
{
	// Notify gmpc that this one's going away
	//gmpc_profile_removed( &plugin, it->profile );

	// Actually free up the memory
/*	g_free( it->profile->id );
	g_free( it->profile->name );
	g_free( it->profile->hostname );
	if( it->profile->password );
		g_free( it->profile->password );
	g_free( it->profile );
	g_free( it );
*/
}

void avahi_del_service( const char *name )
{
/*	ProfileIter* it = profileList;

	while( it && it->profile )
	{
		if( strcmp(it->profile->name, name) == 0 )
			break;
		it = it->next;
	}

	if( it )
	{
		if( profileList == it )
			profileList = it->next;
		else if( it->prev )
			it->prev->next = it->next;

		avahi_free_service( it );
	}

	gmpc_update_profiles_menu();
*/
}

void avahi_del_all_services( void )
{
/*	ProfileIter* it = profileList;

	while( it && it->profile ) {
		ProfileIter* old = it;
		it = it->next;

		g_free( old->profile->id );
		g_free( old->profile->name );
		g_free( old->profile->hostname );
		g_free( old->profile );
		g_free( old );
	}

	profileList = NULL;

	gmpc_update_profiles_menu();
*/
}

/*Profile* avahi_next( Profile* current )
{
	Profile* retval = NULL;
	if( avahi_get_enabled() && current->category == avahi_profiles.category ) {
		if( !current && profileList ) 
			retval = profileList->profile;
		else if( current && current->data ) {
			ProfileIter* it = current->data;
			if( it->next )
				retval = it->next->profile;
		}
	}
	return retval;
}
*/
/*
void avahi_menu_selected(GtkMenuItem* menu, gpointer data)
{
	Profile* element = (Profile*)data;

	fprintf( stderr, "You clicked %s/%s/%d\n", element->name, element->hostname, element->portnumber );
}

int avahi_menu(GtkWidget *menu)
{
	if( avahi_get_enabled() )
	{
		ProfileIter* it;

		// Separator
		gtk_menu_shell_append( GTK_MENU_SHELL(menu), gtk_separator_menu_item_new() );

		// Zeroconf Submenu
		GtkWidget *submenu = gtk_menu_new();
		GtkWidget *subitem = gtk_menu_item_new_with_label( "Zeroconf" );
		gtk_menu_item_set_submenu( GTK_MENU_ITEM(subitem), submenu );
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), subitem);

		if( profileList == NULL ) {
			GtkWidget *item = gtk_image_menu_item_new_with_label( "-- None found --" );
			gtk_widget_set_sensitive(item, FALSE);
			gtk_menu_shell_append(GTK_MENU_SHELL(submenu), item);
		}
		for( it = profileList; it && it->profile; it = it->next ) {
			GtkWidget *item = gtk_image_menu_item_new_with_label( it->profile->name );
			gtk_image_menu_item_set_image( GTK_IMAGE_MENU_ITEM(item),
					gtk_image_new_from_stock( GTK_STOCK_ADD, GTK_ICON_SIZE_MENU ) );
			gtk_menu_shell_append(GTK_MENU_SHELL(submenu), item);
			g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(avahi_menu_selected), it->profile);
		}
	}
	return 1;
}
*/

/**
 * Preferences
 */
static void avahi_profiles_enable_toggle(GtkWidget *wid)
{
	avahi_set_enabled( gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(wid)) );
}

static const char* avahi_get_browse_domain(void)
{
	static char value[128];
	gchar *def = (avahi_get_default_domain());
	if( !def )
		def = "local";
	char* tmpval = cfg_get_single_value_as_string_with_default(config, "avahi-profiles", "domain", def);
	strncpy( value, tmpval, 128 );
	value[127] = '\0';
	cfg_free_string( tmpval );
	return value;
}

static void avahi_profiles_domain_applied(GtkWidget *button, GtkWidget *entry)
{
	const char *str = gtk_entry_get_text(GTK_ENTRY(entry));
	if(str && strcmp( str, avahi_get_browse_domain() ) != 0)
	{
		if( avahi_is_valid_domain_name( str ) ) {
			cfg_set_single_value_as_string(config, "avahi-profiles", "domain",(char *)str);
			fprintf( stderr, "Searching domain '%s'\n", str );
			// Start browsing the new domain
			avahi_domain_changed();
		} else {
			// TODO: Popup an error and set the text back maybe?
			//fprintf( stderr, "Domain '%s' is not valid\n", str );
			gtk_entry_set_text(GTK_ENTRY(entry), avahi_get_browse_domain() );
		}
	}
	gtk_widget_set_sensitive( button, FALSE );
}

static void avahi_profiles_domain_changed(GtkWidget *entry, GtkWidget *button)
{
	const char *str = gtk_entry_get_text(GTK_ENTRY(entry));
	if(str && strcmp( str, avahi_get_browse_domain() ) != 0) {
		// Activate the "apply" button
		gtk_widget_set_sensitive( button, TRUE );
	} else {
		// deactivate the "apply" button
		gtk_widget_set_sensitive( button, FALSE );
	}
}

void avahi_pref_destroy(GtkWidget *container)
{
	gtk_container_remove(GTK_CONTAINER(container), pref_vbox);
}

void avahi_pref_construct(GtkWidget *container)
{
	GtkWidget *entry_hbox = gtk_hbox_new(FALSE,3);
	GtkWidget *enable = gtk_check_button_new_with_mnemonic("_Enable zeroconf profile scanning");
	GtkWidget *entry = gtk_entry_new();
	GtkWidget *apply = gtk_button_new_from_stock( GTK_STOCK_APPLY );
	pref_vbox = gtk_vbox_new(FALSE,6);

	// An 'enabled' checkbox
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(enable), avahi_get_enabled());
	g_signal_connect(G_OBJECT(enable), "toggled", G_CALLBACK(avahi_profiles_enable_toggle), NULL);

	// The domain to browse
	gtk_entry_set_text(GTK_ENTRY(entry), avahi_get_browse_domain());

	// On change of the domain box
	g_signal_connect(G_OBJECT(entry), "changed", G_CALLBACK(avahi_profiles_domain_changed), apply);

	// The "Apply" button
	gtk_widget_set_sensitive( apply, FALSE );
	g_signal_connect(G_OBJECT(apply), "clicked", G_CALLBACK(avahi_profiles_domain_applied), entry);

	// Put the entry and label in the hbox:
	gtk_box_pack_start(GTK_BOX(entry_hbox), gtk_label_new("Search Domain:"), FALSE, FALSE,0);
	gtk_box_pack_start(GTK_BOX(entry_hbox), entry, FALSE, FALSE,0);
	gtk_box_pack_start(GTK_BOX(entry_hbox), apply, FALSE, FALSE,0);

	// Put it all in the vbox:
	gtk_box_pack_start(GTK_BOX(pref_vbox), enable, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(pref_vbox), entry_hbox, FALSE, FALSE,0);

	gtk_container_add(GTK_CONTAINER(container), pref_vbox);
	gtk_widget_show_all(container);
}

/**** Everything below here is all avahi api intergration stuff: ****/

static AvahiClient *client = NULL;
static AvahiServiceBrowser *browser = NULL;

static void resolve_callback(
	AvahiServiceResolver *r,
	AVAHI_GCC_UNUSED AvahiIfIndex interface,
	AVAHI_GCC_UNUSED AvahiProtocol protocol,
	AvahiResolverEvent event,
	const char *name,
	const char *type,
	const char *domain,
	const char *host_name,
	const AvahiAddress *address,
	uint16_t port,
	AvahiStringList *txt,
	AvahiLookupResultFlags flags,
	AVAHI_GCC_UNUSED void* userdata) {

	assert(r);

	/* Called whenever a service has been resolved successfully or timed out */
	printf("resolved: name:%s type:%s domain:%s host_name:%s\n", name,type,domain,host_name);
	switch (event) {
		case AVAHI_RESOLVER_FAILURE:
			fprintf(stderr, "(Resolver) Failed to resolve service '%s' of type '%s' in domain '%s': %s\n", name, type, domain, avahi_strerror(avahi_client_errno(avahi_service_resolver_get_client(r))));
			break;

		case AVAHI_RESOLVER_FOUND: {
			char a[AVAHI_ADDRESS_STR_MAX];
			avahi_address_snprint(a, sizeof(a), address);
			printf("a: %s:%s:%i\n", name,a, port);	
			avahi_add_service( name, a, port );
		}
	}

	avahi_service_resolver_free(r);
}

static void browse_callback(
	AvahiServiceBrowser *b,
	AvahiIfIndex interface,
	AvahiProtocol protocol,
	AvahiBrowserEvent event,
	const char *name,
	const char *type,
	const char *domain,
	AVAHI_GCC_UNUSED AvahiLookupResultFlags flags,
	void* userdata) {
	
	AvahiClient *c = userdata;
	assert(b);
	printf("browser callback: name:%s type:%s domain:%s\n",name,type,domain);
	/* Called whenever a new services becomes available on the LAN or is removed from the LAN */

	switch (event) {
		case AVAHI_BROWSER_FAILURE:
			
			fprintf(stderr, "(Browser) %s\n", avahi_strerror(avahi_client_errno(avahi_service_browser_get_client(b))));
			//avahi_del_all_services();
			return;

		case AVAHI_BROWSER_NEW:
			/* We ignore the returned resolver object. In the callback
			   function we free it. If the server is terminated before
			   the callback function is called the server will free
			   the resolver for us. */

			if (!(avahi_service_resolver_new(c, interface, protocol, name, type, domain, AVAHI_PROTO_UNSPEC, 0, resolve_callback, c)))
				fprintf(stderr, "Failed to resolve service '%s': %s\n", name, avahi_strerror(avahi_client_errno(c)));
			
			break;

		case AVAHI_BROWSER_REMOVE:
			avahi_del_service( name );
			break;

		case AVAHI_BROWSER_ALL_FOR_NOW:
		case AVAHI_BROWSER_CACHE_EXHAUSTED:
			//fprintf(stderr, "(Browser) %s\n", event == AVAHI_BROWSER_CACHE_EXHAUSTED ? "CACHE_EXHAUSTED" : "ALL_FOR_NOW");
			break;
	}
}

static void client_callback(AvahiClient *c, AvahiClientState state, AVAHI_GCC_UNUSED void * userdata) {
	assert(c);
	printf("client callback\n");
	/* Called whenever the client or server state changes */

	if (state == AVAHI_CLIENT_FAILURE) {
		fprintf(stderr, "Server connection failure: %s\n", avahi_strerror(avahi_client_errno(c)));
		// TODO: Maybe try to reconnect here?
	}
}

static void avahi_domain_changed(void)
{
	if( browser ) {
		avahi_service_browser_free(browser);
		avahi_del_all_services();
	}

	browser = avahi_service_browser_new(client, AVAHI_IF_UNSPEC, AVAHI_PROTO_UNSPEC,
			SERVICE_TYPE, avahi_get_browse_domain(), 0, browse_callback, client);
	if( !browser ) {
		fprintf(stderr, "Failed to create service browser for domain %s: %s\n", avahi_get_browse_domain(),
				avahi_strerror(avahi_client_errno(client)));
	}
}

static const char* avahi_get_default_domain()
{
	if( !client )
		return NULL;
	return avahi_client_get_domain_name(client);
}

void avahi_init() {
	AvahiGLibPoll *glib_poll;
	int error;

	glib_poll = avahi_glib_poll_new (NULL, G_PRIORITY_DEFAULT);

	/* Allocate a new client */
	client = avahi_client_new(avahi_glib_poll_get(glib_poll), 0, client_callback, NULL, &error);

	/* Check wether creating the client object succeeded */
	if (!client) {
		fprintf(stderr, "Failed to create client: %s\n", avahi_strerror(error));
		return;
	}
	
	/* Create the service browser */
	avahi_domain_changed();
}
